/*
 * I am the beam correction engine for the 02 feedback type.
 *
 * Macro parameters:
 *   P     PV name prefix (e.g. "17idb:")
 *   A     Auto beam positioner name (e.g. Abp2)
 *   BPPG  Beam positioning permit granter (e.g. ioc23:Bppg)
 */

program autobp02

option +r;

%%#include <string.h>

%%#include "autobpLib.h"

#include "autobpSnl.h"

/* Control variables */

short Go; assign Go to "{P}{A}:Go";
  monitor Go;

double CorrTrigX; assign CorrTrigX to "{P}{A}:CorrTrigX";
  monitor CorrTrigX;
  evflag CorrTrigXChanged;
  sync CorrTrigX CorrTrigXChanged;

double CorrTrigY; assign CorrTrigY to "{P}{A}:CorrTrigY";
  monitor CorrTrigY;
  evflag CorrTrigYChanged;
  sync CorrTrigY CorrTrigYChanged;

double OutPosX; assign OutPosX to "{P}{A}:OutPosX";
double OutPosY; assign OutPosY to "{P}{A}:OutPosY";

double RqsPosX; assign RqsPosX to "{P}{A}:RqsPosX";
double RqsPosY; assign RqsPosY to "{P}{A}:RqsPosY";

short StopX; assign StopX to "{P}{A}:StopX.PROC";
short StopY; assign StopY to "{P}{A}:StopY.PROC";

/* State sets */

ss Main {
  int MovedX;
  int MoveCompletedX;
  int MovedY;
  int MoveCompletedY;

  state initialize {
    when (pvConnectCount() == pvChannelCount()) {
      LogIdx = pvIndex(Log);
    } state initializeAfterDelay
  }

  state initializeAfterDelay {
    when (delay(AUTOBP_INITIALIZE_DELAY)) {
      efClear(AbortRequested);
      efClear(CorrTrigXChanged);
      efClear(CorrTrigYChanged);
    } state waitUntilGo
  }

  state paused {
    entry {
      State = AUTOBP_STATE_PAUSED;
      pvPut(State);
    }

    when (!ExtPauseNeeded) {
    } state waitUntilGo
  }

  state waitUntilGo {
    entry {
      State = AUTOBP_STATE_SUSPENDED;
      pvPut(State);
    }

    when (Go) {
    } state waitUntilCorrectionNeeded

    when (ExtPauseNeeded) {
    } state paused
  }

  state waitUntilCorrectionNeeded {
    entry {
      State = AUTOBP_STATE_MONITORING;
      pvPut(State);
      efClear(CorrTrigXChanged);
      efClear(CorrTrigYChanged);
    }

    when (!Go) {
    } state waitUntilGo

    when (ExtPauseNeeded) {
    } state paused

    when (efTest(CorrTrigXChanged) || efTest(CorrTrigYChanged)) {
    } state correctionNeeded
  }

  state correctionNeeded {
    when () {
      efClear(AbortRequested);
    } state requestPositioningPermit
  }

  state requestPositioningPermit {
    entry {
      State = AUTOBP_STATE_GETTING_PERMIT;
      pvPut(State);
    }

    when () {
      BppgIn = AUTOBP_BPPG_IN_REQUEST;
      pvPut(BppgIn, SYNC);
    } state waitToObtainPositioningPermit
  }

  state waitToObtainPositioningPermit {
    when (BppgOut == AUTOBP_BPPG_OUT_GRANTED) {
    } state doCorrection

    when (BppgOut == AUTOBP_BPPG_OUT_RESET) {
    } state requestPositioningPermit

    when (delay(AUTOBP_OBTAIN_PERMIT_TIMEOUT)) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Timeout waiting for positioning permit");
      }%
    } state didNotWaitToObtainPositioningPermit

    when (!Go) {
    } state didNotWaitToObtainPositioningPermit

    when (efTestAndClear(AbortRequested)) {
    } state didNotWaitToObtainPositioningPermit
  }

  state didNotWaitToObtainPositioningPermit {
    when () {
      BppgIn = AUTOBP_BPPG_IN_RELEASE;
      pvPut(BppgIn);
    } state waitUntilCorrectionNeeded
  }

  state doCorrection {
    entry {
      State = AUTOBP_STATE_CORRECTING;
      pvPut(State);
    }

    when () {
      MovedX = 0;
      MoveCompletedX = 1;
      MovedY = 0;
      MoveCompletedY = 1;

      if (efTestAndClear(CorrTrigXChanged)) {
        Status = pvGet(OutPosX);
        if (Status == AUTOBP_SNCSEQ_PV_STAT_OK) {
          MovedX = 1;
          MoveCompletedX = 0;
          RqsPosX = OutPosX;
          pvPut(RqsPosX, ASYNC);
        } else {
          %{
            autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
                "Failed to get X positioner target");
          }%
        }
      }

      if (efTestAndClear(CorrTrigYChanged)) {
        Status = pvGet(OutPosY);
        if (Status == AUTOBP_SNCSEQ_PV_STAT_OK) {
          MovedY = 1;
          MoveCompletedY = 0;
          RqsPosY = OutPosY;
          pvPut(RqsPosY, ASYNC);
        } else {
          %{
            autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
                "Failed to get Y positioner target");
          }%
        }
      }
    } state waitForMovesToFinish
  }

  state waitForMovesToFinish {
    option -t;

    when (!MoveCompletedX && pvPutComplete(RqsPosX)) {
      MoveCompletedX = 1;
    } state waitForMovesToFinish
    when (!MoveCompletedY && pvPutComplete(RqsPosY)) {
      MoveCompletedY = 1;
    } state waitForMovesToFinish
    when (MoveCompletedX && MoveCompletedY) {
    } state releasePositioningPermit

    when (BppgOut == AUTOBP_BPPG_OUT_REVOKED ||
        BppgOut == AUTOBP_BPPG_OUT_RESET) {
      if (BppgOut == AUTOBP_BPPG_OUT_REVOKED) {
        %{
          autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
              "Positioning permit revoked");
        }%
      }
      if (BppgOut == AUTOBP_BPPG_OUT_RESET) {
        %{
          autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
              "Received BPPG protocol reset");
        }%
      }
    } state didNotWaitForMovesToFinish

    when (delay(AUTOBP_POSITIONING_TIMEOUT)) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Timeout waiting for correction finish");
      }%
    } state didNotWaitForMovesToFinish

    when (efTestAndClear(AbortRequested)) {
    } state didNotWaitForMovesToFinish
  }

  state didNotWaitForMovesToFinish {
    when () {
      if (MovedX) {
        pvPutCancel(RqsPosX);
        StopX = 1;
        pvPut(StopX);
      }
      if (MovedY) {
        pvPutCancel(RqsPosY);
        StopY = 1;
        pvPut(StopY);
      }
    } state releasePositioningPermit
  }

  state releasePositioningPermit {
    when () {
      BppgIn = AUTOBP_BPPG_IN_RELEASE;
      pvPut(BppgIn, SYNC);
    } state waitForReleasePositioningPermitResponse
  }

  state waitForReleasePositioningPermitResponse {
    when (BppgOut == AUTOBP_BPPG_OUT_RELEASE_OK) {
    } state delayAfterCorrection

    when (BppgOut == AUTOBP_BPPG_OUT_RESET) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Received BPPG protocol reset");
      }%
    } state delayAfterCorrection

    when (delay(AUTOBP_RELEASE_PERMIT_RESPONSE_TIMEOUT)) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Timeout waiting for permit release OK");
      }%
    } state delayAfterCorrection

    when (efTestAndClear(AbortRequested)) {
    } state delayAfterCorrection
  }

  state delayAfterCorrection {
    entry {
      if (CorrDly > 0.0) {
        State = AUTOBP_STATE_DELAYING;
        pvPut(State);
      }
    }

    when (delay(CorrDly)) {
    } state afterCorrection

    when (efTestAndClear(AbortRequested)) {
    } state afterCorrection
  }

  state afterCorrection {
    when () {
      efClear(CorrTrigXChanged);
      efClear(CorrTrigYChanged);
    } state waitUntilCorrectionNeeded
  }
}
