/*
 * I am the beam correction engine for the 04 feedback type.
 *
 * Macro parameters:
 *   P     PV name prefix (e.g. "17ida:")
 *   A     Auto beam positioner name (e.g. Abp1)
 *   BPPG  Beam positioning permit granter (e.g. ioc23:Bppg)
 */

program autobp04

option +r;

%%#include <string.h>

%%#include "autobpLib.h"

#include "autobpSnl.h"
#include "autobpInterruptSnl.h"

/* General constants */

#define AUTOBP_SCAN_MAXPTS  1000
#define AUTOBP_SCAN_CNT_MAX   10

/* Control variables */

short Go; assign Go to "{P}{A}:Go";
  monitor Go;

double CorrTrigX; assign CorrTrigX to "{P}{A}:CorrTrigX";
  monitor CorrTrigX;
  evflag CorrTrigXChanged;
  sync CorrTrigX CorrTrigXChanged;

double RqsPosX; assign RqsPosX to "{P}{A}:RqsPosX";
double ActPosX; assign ActPosX to "{P}{A}:ActPosX";
char ActPosProcX; assign ActPosProcX to "{P}{A}:ActPosX.PROC";

short StopX; assign StopX to "{P}{A}:StopX.PROC";

short CorrectionDone; assign CorrectionDone to "{P}{A}:CorrectionDone.PROC";

/* Dependent auto beam positioner pausing */

short DepExtPause; assign DepExtPause to "{P}{A}:DepExtPause";

/* Scan variables */

short ScanExecX; assign ScanExecX to "{P}{A}:ScanX.EXSC";
double ScanPosDataX[AUTOBP_SCAN_MAXPTS];
  assign ScanPosDataX to "{P}{A}:ScanX.P1RA";
float ScanPosSettleTimeX; assign ScanPosSettleTimeX to "{P}{A}:ScanX.PDLY";
  monitor ScanPosSettleTimeX;
float ScanDetDataX[AUTOBP_SCAN_MAXPTS];
  assign ScanDetDataX to "{P}{A}:ScanX.D01DA";
int ScanNumPtsX; assign ScanNumPtsX to "{P}{A}:ScanX.NPTS";
char ScanAlertX; assign ScanAlertX to "{P}{A}:ScanX.ALRT";
short ScanBusyX; assign ScanBusyX to "{P}{A}:ScanX.BUSY"; monitor ScanBusyX;

int PkStrtPtX; assign PkStrtPtX to "{P}{A}:PkStrtPtX";

int ScanCnt; assign ScanCnt to "{P}{A}:04ScanCntX";

double BestDetX; assign BestDetX to "{P}{A}:ScanBestDetX";

int DoOptimizationScanHadError;
int MaxAtEdge;
int ScanFailed;
double OrigPosX;
double BestPosX;
int StartIdx;
int MaxIdx;
int LostBppgPermitDuringOptimizationScan;

/* Support variables */

int Status2;
int Status3;
int Status4;
int Status5;

/* State sets */

ss Main {
  int MovedX;

  state initialize {
    when (pvConnectCount() == pvChannelCount()) {
      LogIdx = pvIndex(Log);
    } state initializeAfterDelay
  }

  state initializeAfterDelay {
    when (delay(AUTOBP_INITIALIZE_DELAY)) {
      efClear(AbortRequested);
      efClear(InterruptRequested);
      efClear(CorrTrigXChanged);
    } state waitUntilGo
  }

  state paused {
    entry {
      State = AUTOBP_STATE_PAUSED;
      pvPut(State);
    }

    when (!ExtPauseNeeded) {
    } state waitUntilGo
  }

  state waitUntilGo {
    entry {
      State = AUTOBP_STATE_SUSPENDED;
      pvPut(State);
    }

    when (Go) {
    } state waitUntilCorrectionNeeded

    when (ExtPauseNeeded) {
    } state paused
  }

  state waitUntilCorrectionNeeded {
    entry {
      State = AUTOBP_STATE_MONITORING;
      pvPut(State);
      efClear(CorrTrigXChanged);
    }

    when (!Go) {
    } state waitUntilGo

    when (ExtPauseNeeded) {
    } state paused

    when (efTest(CorrTrigXChanged)) {
    } state correctionNeeded
  }

  state correctionNeeded {
    when () {
      efClear(AbortRequested);
    } state pauseDependentAbp
  }

  state pauseDependentAbp {
    entry {
      State = AUTOBP_STATE_PAUSING_DEP_ABP;
      pvPut(State);
    }

    when () {
      DepExtPause = 1;
      pvPut(DepExtPause, ASYNC);
    } state waitForDependentAbpToPause
  }

  state waitForDependentAbpToPause {
    when (pvPutComplete(DepExtPause)) {
    } state requestPositioningPermit

    when (delay(AUTOBP_DEP_ABP_PAUSE_TIMEOUT)) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Timeout waiting for dep ABP to pause");
      }%
    } state didNotWaitForDependentAbpToPause

    when (efTestAndClear(AbortRequested)) {
    } state didNotWaitForDependentAbpToPause
  }

  state didNotWaitForDependentAbpToPause {
    when () {
      pvPutCancel(DepExtPause);
      DepExtPause = 0;
      pvPut(DepExtPause);
    } state waitUntilCorrectionNeeded
  }

  state requestPositioningPermit {
    entry {
      State = AUTOBP_STATE_GETTING_PERMIT;
      pvPut(State);
    }

    when () {
      BppgIn = AUTOBP_BPPG_IN_REQUEST;
      pvPut(BppgIn, SYNC);
    } state waitToObtainPositioningPermit
  }

  state waitToObtainPositioningPermit {
    when (BppgOut == AUTOBP_BPPG_OUT_GRANTED) {
    } state beforeOptimizationScan

    when (BppgOut == AUTOBP_BPPG_OUT_RESET) {
    } state requestPositioningPermit

    when (delay(AUTOBP_OBTAIN_PERMIT_TIMEOUT)) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Timeout waiting for positioning permit");
      }%
    } state didNotWaitToObtainPositioningPermit

    when (!Go) {
    } state didNotWaitToObtainPositioningPermit

    when (efTestAndClear(AbortRequested)) {
    } state didNotWaitToObtainPositioningPermit
  }

  state didNotWaitToObtainPositioningPermit {
    when () {
      BppgIn = AUTOBP_BPPG_IN_RELEASE;
      pvPut(BppgIn);
      DepExtPause = 0;
      pvPut(DepExtPause);
    } state waitUntilCorrectionNeeded
  }

  state beforeOptimizationScan {
    when () {
      ScanCnt = 0;
      pvPut(ScanCnt);
      efClear(InterruptRequested);
    } state doOptimizationScan
  }

  state doOptimizationScan {
    entry {
      State = AUTOBP_STATE_SCANNING;
      pvPut(State);
      LostBppgPermitDuringOptimizationScan = 0;
    }

    when (ScanCnt > AUTOBP_SCAN_CNT_MAX) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Exceeded repeat scan limit");
      }%
    } state releasePositioningPermit

    when (ScanCnt <= AUTOBP_SCAN_CNT_MAX) {
      ScanCnt += 1;
      pvPut(ScanCnt);
      ScanFailed = 1;
      MaxAtEdge = 0;
      DoOptimizationScanHadError = 1;
      ActPosProcX = 1;
      Status = pvPut(ActPosProcX, SYNC);
      Status2 = pvGet(ActPosX);
      if (Status == AUTOBP_SNCSEQ_PV_STAT_OK &&
          Status2 == AUTOBP_SNCSEQ_PV_STAT_OK) {
        DoOptimizationScanHadError = 0;
        OrigPosX = ActPosX;
        ScanExecX = AUTOBP_SCAN_EXSC_START;
        pvPut(ScanExecX, ASYNC);
      } else {
        %{
          autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
              "Failed to get positioner original pos.");
        }%
      }
    } state checkForDoOptimizationScanError
  }

  state checkForDoOptimizationScanError {
    when (!DoOptimizationScanHadError) {
    } state waitForOptimizationScanToFinish

    when (DoOptimizationScanHadError) {
    } state releasePositioningPermit
  }

  state waitForOptimizationScanToFinish {
    when (pvPutComplete(ScanExecX)) {
      Status = pvGet(ScanAlertX);
      Status2 = pvGet(ScanPosDataX);
      Status3 = pvGet(ScanDetDataX);
      Status4 = pvGet(ScanNumPtsX);
      Status5 = pvGet(PkStrtPtX);
      if (Status == AUTOBP_SNCSEQ_PV_STAT_OK &&
          Status2 == AUTOBP_SNCSEQ_PV_STAT_OK &&
          Status3 == AUTOBP_SNCSEQ_PV_STAT_OK &&
          Status4 == AUTOBP_SNCSEQ_PV_STAT_OK &&
          Status5 == AUTOBP_SNCSEQ_PV_STAT_OK &&
          ScanAlertX == AUTOBP_SCAN_ALERT_NO) {
        if (ScanNumPtsX > pvCount(ScanDetDataX))
          ScanNumPtsX = pvCount(ScanDetDataX);
        StartIdx = PkStrtPtX - 1;
        if (StartIdx < 0) StartIdx = 0;
        if (StartIdx > ScanNumPtsX - 1) StartIdx = ScanNumPtsX - 1;
        %{
          pVar->MaxIdx = autobpFindMaxFloat(pVar->ScanDetDataX,
              pVar->StartIdx, pVar->ScanNumPtsX);
        }%
        if (MaxIdx >= StartIdx) {
          ScanFailed = 0;
          BestPosX = ScanPosDataX[MaxIdx];
          BestDetX = ScanDetDataX[MaxIdx];
          pvPut(BestDetX);
          if (MaxIdx == StartIdx || MaxIdx == ScanNumPtsX - 1) MaxAtEdge = 1;
        } else {
          %{
            autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
                "No maximum found in scan data");
          }%
        }
      } else {
        if (Status != AUTOBP_SNCSEQ_PV_STAT_OK) {
          %{
            autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
                "Failed to get scan alert value");
          }%
        } else if (ScanAlertX != AUTOBP_SCAN_ALERT_NO) {
          %{
            autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
                "Scan failed with alert");
          }%
        }
        if (Status2 != AUTOBP_SNCSEQ_PV_STAT_OK) {
          %{
            autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
                "Failed to get scan position values");
          }%
        }
        if (Status3 != AUTOBP_SNCSEQ_PV_STAT_OK) {
          %{
            autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
                "Failed to get scan detector values");
          }%
        }
        if (Status4 != AUTOBP_SNCSEQ_PV_STAT_OK) {
          %{
            autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
                "Failed to get scan number of points");
          }%
        }
        if (Status5 != AUTOBP_SNCSEQ_PV_STAT_OK) {
          %{
            autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
                "Failed to get peak search start point");
          }%
        }
      }
    } state checkForWaitForOptimizationScanToFinishError

    when (BppgOut == AUTOBP_BPPG_OUT_REVOKED ||
        BppgOut == AUTOBP_BPPG_OUT_RESET) {
      LostBppgPermitDuringOptimizationScan = 1;
      if (BppgOut == AUTOBP_BPPG_OUT_REVOKED) {
        %{
          autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
              "Positioning permit revoked");
        }%
      }
      if (BppgOut == AUTOBP_BPPG_OUT_RESET) {
        %{
          autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
              "Received BPPG protocol reset");
        }%
      }
    } state didNotWaitForOptimizationScanToFinish

    when (delay(AUTOBP_SCANNING_TIMEOUT)) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Timeout waiting for scan to finish");
      }%
    } state didNotWaitForOptimizationScanToFinish

    when (efTestAndClear(AbortRequested)) {
    } state didNotWaitForOptimizationScanToFinish

    when (efTestAndClear(InterruptRequested)) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Interrupted");
      }%
    } state didNotWaitForOptimizationScanToFinish
  }

  state checkForWaitForOptimizationScanToFinishError {
    when (!ScanFailed) {
    } state doCorrection

    when (ScanFailed) {
    } state releasePositioningPermit
  }

  state didNotWaitForOptimizationScanToFinish {
    when () {
      pvPutCancel(ScanExecX);
      ScanExecX = AUTOBP_SCAN_EXSC_ABORT;
      pvPut(ScanExecX);
      StopX = 1;
      pvPut(StopX);
    } state waitForOptimizationScanToAbort
  }

  state waitForOptimizationScanToAbort {
    when (ScanBusyX == 0) {
    } state moveToOriginalPosition
    when (delay(AUTOBP_SCAN_ABORT_TIMEOUT)) {
      ScanExecX = AUTOBP_SCAN_EXSC_ABORT;
      pvPut(ScanExecX);
      pvFlush();
      pvPut(ScanExecX);
    } state delayAfterForcedOptimizationScanAbort
  }

  state delayAfterForcedOptimizationScanAbort {
    when (delay(AUTOBP_SCAN_ABORT_FORCED_DELAY)) {
    } state moveToOriginalPosition
  }

  state moveToOriginalPosition {
    when (LostBppgPermitDuringOptimizationScan) {
    } state releasePositioningPermit
    when (!LostBppgPermitDuringOptimizationScan) {
      RqsPosX = OrigPosX;
      pvPut(RqsPosX, ASYNC);
    } state waitForMoveToOriginalPositionToFinish
  }

  state waitForMoveToOriginalPositionToFinish {
    when (pvPutComplete(RqsPosX)) {
    } state releasePositioningPermit

    when (BppgOut == AUTOBP_BPPG_OUT_REVOKED ||
        BppgOut == AUTOBP_BPPG_OUT_RESET) {
      if (BppgOut == AUTOBP_BPPG_OUT_REVOKED) {
        %{
          autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
              "Positioning permit revoked");
        }%
      }
      if (BppgOut == AUTOBP_BPPG_OUT_RESET) {
        %{
          autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
              "Received BPPG protocol reset");
        }%
      }
    } state didNotWaitForMoveToOriginalPositionToFinish

    when (delay(AUTOBP_POSITIONING_TIMEOUT)) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Timeout waiting for move to orig. pos.");
      }%
    } state didNotWaitForMoveToOriginalPositionToFinish

    when (efTestAndClear(AbortRequested)) {
    } state didNotWaitForMoveToOriginalPositionToFinish
  }

  state didNotWaitForMoveToOriginalPositionToFinish {
    when () {
      pvPutCancel(RqsPosX);
      StopX = 1;
      pvPut(StopX);
    } state releasePositioningPermit
  }

  state doCorrection {
    entry {
      State = AUTOBP_STATE_CORRECTING;
      pvPut(State);
    }

    when () {
      MovedX = 0;

      if (efTest(CorrTrigXChanged)) {
        MovedX = 1;
        RqsPosX = BestPosX;
        pvPut(RqsPosX, ASYNC);
      }
    } state waitForCorrectionMoveToFinish
  }

  state waitForCorrectionMoveToFinish {
    when (!MovedX) {
    } state correctionFinished
    when (MovedX && pvPutComplete(RqsPosX)) {
    } state waitForCorrectionMoveSettleTime

    when (BppgOut == AUTOBP_BPPG_OUT_REVOKED ||
        BppgOut == AUTOBP_BPPG_OUT_RESET) {
      if (BppgOut == AUTOBP_BPPG_OUT_REVOKED) {
        %{
          autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
              "Positioning permit revoked");
        }%
      }
      if (BppgOut == AUTOBP_BPPG_OUT_RESET) {
        %{
          autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
              "Received BPPG protocol reset");
        }%
      }
    } state didNotWaitForCorrectionMoveToFinish

    when (delay(AUTOBP_POSITIONING_TIMEOUT)) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Timeout waiting for correction to finish");
      }%
    } state didNotWaitForCorrectionMoveToFinish

    when (efTestAndClear(AbortRequested)) {
    } state didNotWaitForCorrectionMoveToFinish
  }

  state waitForCorrectionMoveSettleTime {
    when (delay(ScanPosSettleTimeX)) {
    } state correctionFinished
  }

  state correctionFinished {
    entry {
      CorrectionDone = 1;
      pvPut(CorrectionDone, SYNC);
    }

    when (MaxAtEdge) {
    } state delayBeforeRepeatingOptimizationScan

    when (!MaxAtEdge) {
    } state releasePositioningPermit
  }

  state delayBeforeRepeatingOptimizationScan {
    when (delay(AUTOBP_REPEAT_OPTIMIZATION_SCAN_DELAY)) {
    } state doOptimizationScan
  }

  state didNotWaitForCorrectionMoveToFinish {
    when () {
      if (MovedX) {
        pvPutCancel(RqsPosX);
        StopX = 1;
        pvPut(StopX);
      }
    } state releasePositioningPermit
  }

  state releasePositioningPermit {
    when () {
      BppgIn = AUTOBP_BPPG_IN_RELEASE;
      pvPut(BppgIn, SYNC);
    } state waitForReleasePositioningPermitResponse
  }

  state waitForReleasePositioningPermitResponse {
    when (BppgOut == AUTOBP_BPPG_OUT_RELEASE_OK) {
    } state unpauseDependentAbp

    when (BppgOut == AUTOBP_BPPG_OUT_RESET) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Received BPPG protocol reset");
      }%
    } state unpauseDependentAbp

    when (delay(AUTOBP_RELEASE_PERMIT_RESPONSE_TIMEOUT)) {
      %{
        autobpLog(ssId, pVar->LogIdx, pVar->Log, sizeof(pVar->Log),
            "Timeout waiting for permit release OK");
      }%
    } state unpauseDependentAbp

    when (efTestAndClear(AbortRequested)) {
    } state unpauseDependentAbp
  }

  state unpauseDependentAbp {
    when () {
      DepExtPause = 0;
      pvPut(DepExtPause, SYNC);
    } state delayAfterCorrection
  }

  state delayAfterCorrection {
    entry {
      if (CorrDly > 0.0) {
        State = AUTOBP_STATE_DELAYING;
        pvPut(State);
      }
    }

    when (delay(CorrDly)) {
    } state afterCorrection

    when (efTestAndClear(AbortRequested)) {
    } state afterCorrection
  }

  state afterCorrection {
    when () {
      efClear(CorrTrigXChanged);
    } state waitUntilCorrectionNeeded
  }
}
